/*   Gtk+ version of the newt library
 *
 *   Copyright (C) 1996-1999  Red Hat Software
 *   Copyright (C) 1999-2000  O'ksi'D
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *       O'ksi'D <nickasil@linuxbox.com>
 *       Jean-Marc Lienher
 *       Rue de la Cheminee 1
 *       CH-2065 Savagnier
 *       Switzerland
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "config.h"

#include "newt.h"
#include "newt_pr.h"

struct Window {
    int height, width, top, left;
    short * buffer;
    char * title;
    GtkWidget* widget;
    GtkWidget* bg_widget;
};

struct keymap {
    char * str;
    int code;
    char * tc;
};

static struct Window windowStack[20];

/* gNewt special variable */
struct Window * currentWindow = NULL;
int pressedKey = 0;
static  GtkWidget *gRootWindow = NULL;
GtkWidget* currentParent = NULL;
int FontSizeW = 640 / 80;
int FontSizeH = 480 / 24;
int ScreenCols = 80;
int ScreenRows = 24;
int ScreenW = 640;
int ScreenH = 480;
/* gNewt end */

static char * helplineStack[20];
static char ** currentHelpline = NULL;
static GtkWidget * currentHelpWidget = NULL;

static int cursorRow, cursorCol;
static int needResize;
static int threeDbox = 0;

static const char * defaultHelpLine = 
"  <Tab>/<Alt-Tab> between elements   |  <Space> selects   |  <F12> next screen"
;

struct newtColors newtDefaultColorPalette = {
	"white", "blue", 			/* root fg, bg */     
	"black", "lightgray",			/* border fg, bg */
	"black", "lightgray",			/* window fg, bg */
	"white", "black",			/* shadow fg, bg */
	"red", "lightgray",			/* title fg, bg */
	"lightgray", "red",			/* button fg, bg */
	"red", "lightgray",			/* active button fg, bg */
	"yellow", "blue",			/* checkbox fg, bg */
	"blue", "brown",			/* active checkbox fg, bg */
	"yellow", "blue",			/* entry box fg, bg */
	"blue", "lightgray",			/* label fg, bg */
	"black", "lightgray",			/* listbox fg, bg */
	"yellow", "blue",			/* active listbox fg, bg */
	"black", "lightgray",			/* textbox fg, bg */
	"lightgray", "black",			/* active textbox fg, bg */
	"white", "blue",			/* help line */
	"yellow", "blue",			/* root text */
	"blue",					/* scale full */
	"red",					/* scale empty */
	"blue", "lightgray",			/* disabled entry fg, bg */
	"white", "blue",			/* compact button fg, bg */
	"yellow", "red",			/* active & sel listbox */
	"black", "brown",			/* selected listbox */
};

struct newtColors * globalColors;

struct {
  GdkColor color;  
  gchar *name;
} gtk_colors[] = {
 { { 0, 0x0000, 0x0000, 0x0000 }, "black" },
 { { 0, 0xFFFF, 0xFFFF, 0xFFFF }, "white" },
 { { 0, 0xFFFF, 0x0000, 0x0000 }, "red" },
 { { 0, 0x0000, 0xFFFF, 0x0000 }, "green" },
 { { 0, 0x0000, 0x0000, 0xFFFF }, "blue" },
 { { 0, 0x0000, 0xFFFF, 0xFFFF }, "cyan" },
 { { 0, 0xFFFF, 0x0000, 0xFFFF }, "magenta" },
 { { 0, 0xFFFF, 0xFFFF, 0x0000 }, "yellow" }, 
/* { { 0, 0x8000, 0x8000, 0x8000 }, "gray" }, */
 { { 0, 0xC000, 0xC000, 0xC000 }, "lightgray" },
 { { 0, 0x8000, 0x8000, 0x0000 }, "brown" },
 { { 0, 0x8000, 0x0000, 0x8000 }, "darkmagenta" },
 { { 0, 0x0000, 0x8000, 0x8000 }, "darkcyan" },
 { { 0, 0x0000, 0x8000, 0x0000 }, "darkgreen" },
 { { 0, 0x0000, 0x0000, 0x8000 }, "darkblue" },
 { { 0, 0xFFFF, 0x8000, 0x0000 }, "orange" },
 { { 0, 0xFFFF, 0xFFFF, 0x8000 }, "mediumgreen" },
 { { 0, 0, 0, 0}, NULL },
};

static const struct keymap keymap[] = {
	{ "\033OA", 		NEWT_KEY_UP, 		"kh" },
	{ "\033[A", 		NEWT_KEY_UP, 		"ku" },
	{ "\033OB", 		NEWT_KEY_DOWN, 		"kd" },
	{ "\033[B", 		NEWT_KEY_DOWN, 		"kd" },
	{ "\033[C", 		NEWT_KEY_RIGHT, 	"kr" },
	{ "\033OC", 		NEWT_KEY_RIGHT, 	"kr" },
	{ "\033[D", 		NEWT_KEY_LEFT, 		"kl" },
	{ "\033OD", 		NEWT_KEY_LEFT, 		"kl" },
	{ "\033[H",		NEWT_KEY_HOME, 		"kh" },
	{ "\033[1~",		NEWT_KEY_HOME, 		"kh" },
	{ "\033Ow",		NEWT_KEY_END, 		"kH" },
	{ "\033[4~",		NEWT_KEY_END, 		"kH" },

	{ "\033[3~",		NEWT_KEY_DELETE,	"kl" },

	{ "\033\t",		NEWT_KEY_UNTAB,		NULL },

	{ "\033[5~",		NEWT_KEY_PGUP,		NULL },
	{ "\033[6~",		NEWT_KEY_PGDN,		NULL },
	{ "\033V",		NEWT_KEY_PGUP, 		"kH" },
	{ "\033v",		NEWT_KEY_PGUP, 		"kH" },

	{ "\033[[A",		NEWT_KEY_F1,		NULL },
	{ "\033[[B",		NEWT_KEY_F2,		NULL },
	{ "\033[[C",		NEWT_KEY_F3,		NULL },
	{ "\033[[D",		NEWT_KEY_F4,		NULL },
	{ "\033[[E",		NEWT_KEY_F5,		NULL },

	{ "\033[11~",		NEWT_KEY_F1,		NULL },
	{ "\033[12~",		NEWT_KEY_F2,		NULL },
	{ "\033[13~",		NEWT_KEY_F3,		NULL },
	{ "\033[14~",		NEWT_KEY_F4,		NULL },
	{ "\033[15~",		NEWT_KEY_F5,		NULL },
	{ "\033[17~",		NEWT_KEY_F6,		NULL },
	{ "\033[18~",		NEWT_KEY_F7,		NULL },
	{ "\033[19~",		NEWT_KEY_F8,		NULL },
	{ "\033[20~",		NEWT_KEY_F9,		NULL },
	{ "\033[21~",		NEWT_KEY_F10,		NULL },
	{ "\033[23~",		NEWT_KEY_F11,		NULL },
	{ "\033[24~",		NEWT_KEY_F12,		NULL },

	{ NULL, 	0, 			NULL },	/* LEAVE this one */
};

static char keyPrefix = '\033';

static const char * version = "gNewt : GTK wrapper for Newt version " VERSION
			" - (C) 1996 Red Hat Software - (C) 1999 O'ksi'D. "
		        "Redistributable under the term of the Library "
		        "GNU Public License. "
			"gNewt : Adapted by O'ksi'D. "
			"Newt : Written by Erik Troan\n";

static newtSuspendCallback suspendCallback = NULL;
static void * suspendCallbackData = NULL;

void newtSetSuspendCallback(newtSuspendCallback cb, void * data) {
    suspendCallback = cb;
    suspendCallbackData = data;
}


static void handleSigwinch(int signum) {
    needResize = 1;
}

static int getkeyInterruptHook(void) {
    return -1;
}

GdkColor getColorByName (gchar *name) 
{
	int i = 0;
	while (gtk_colors[i].name && strcmp(name, gtk_colors[i].name)) i++;
	return gtk_colors[i].color;	
}

/*
 * threeDbox = 0 - no 3D boxes
 *           = 1 - 3D raised textboxes, sunken listboxes
 */
void newtSetThreeD(int val) 
{
        threeDbox = val;
}

void newtFlushInput(void) {
}

void newtRefresh(void) {
	while (gtk_events_pending ()) gtk_main_iteration ();
}

void newtSuspend(void) {
}

void newtResume(void) {
}

void newtCls(void) {
}

void newtResizeScreen(int redraw) {
	if (redraw) newtRefresh ();
}

static gint keyPressedCallback (gpointer unused, GdkEventKey *event ) {
    /* printf("%s : %c\n", event->string, event->keyval); */
    switch (event->keyval) {
	case 65307: 
		pressedKey = NEWT_KEY_SUSPEND; 
		break;
	case 65481: 
		pressedKey = NEWT_KEY_F12; break;
        case 65480: 
                pressedKey = NEWT_KEY_F11; break;
        case 65479: 
                pressedKey = NEWT_KEY_F10; break;
        case 65478: 
                pressedKey = NEWT_KEY_F9; break;
        case 65477: 
                pressedKey = NEWT_KEY_F8; break;
        case 65476: 
                pressedKey = NEWT_KEY_F7; break;
        case 65475: 
                pressedKey = NEWT_KEY_F6; break;
        case 65474: 
                pressedKey = NEWT_KEY_F5; break;
        case 65473:
                pressedKey = NEWT_KEY_F4; break;
        case 65472:
                pressedKey = NEWT_KEY_F3; break;
        case 65471:
                pressedKey = NEWT_KEY_F2; break;
        case 65470:
                pressedKey = NEWT_KEY_F1; break;
	default: 
		return TRUE;
    }	
    gtk_main_quit();
    return TRUE;
}

int newtInit(void) {
    GtkWidget *win;
    GtkStyle *style; 
    GtkWidget *scrolledwindow;
    char *var = NULL;
/*   GdkIC *ic; 
 *   GdkICAttr *ic_attr;
 */
 
    /* use the version variable just to be sure it gets included */
    strlen (version);
    
    (void) keyPrefix; /* to avoid compiler warning*/
    
    globalColors = malloc (sizeof(struct newtColors));


    newtSetColors (newtDefaultColorPalette);
    handleSigwinch (0);
    getkeyInterruptHook ();

    /* XIM protocol compliant */
    gtk_set_locale();
    gtk_init (NULL,NULL);

    /* set a good default color contrast for 16 colors display */
    style = gtk_widget_get_default_style();
    style->bg[0] = getColorByName ("lightgray");

    win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    style = gtk_rc_get_style(win);
    if (!style) style = gtk_widget_get_style (win);
    FontSizeH = style->font->ascent + style->font->descent + 2;
    FontSizeW = gdk_char_measure (style->font, '_'); 
   
    var = getenv ("GNEWT_FONT_SIZE");
    if (var) {
	char **size;
        size = g_strsplit (var, "x",-1);
	if (size[0]) FontSizeH = atoi (size[0]);
	if (size[1]) FontSizeW = atoi (size[1]);
    } 	
		
    if (FontSizeH < 14) FontSizeH = 14;	
    if (FontSizeW < 7) FontSizeW = 7;
    ScreenW = FontSizeW * ScreenCols;
    ScreenH = FontSizeH * ScreenRows;

		
    gtk_signal_connect_object (GTK_OBJECT (win), "key_press_event",
        GTK_SIGNAL_FUNC (keyPressedCallback), NULL);
    gtk_widget_set_events (win, GDK_KEY_PRESS_MASK);
    gtk_widget_set_usize (win, ScreenW, ScreenH);

    GTK_WINDOW(win)->allow_shrink = TRUE;
    GTK_WINDOW(win)->title = "gNewt window";
    GTK_WINDOW(win)->allow_grow = FALSE;
    /* gtk_widget_set_uposition (win, 0,0); */
    gtk_container_set_border_width (GTK_CONTAINER (win), 0);
    gtk_widget_show (win);

/*    ic_attr = gdk_ic_attr_new ();
 *   ic_attr->style = gdk_im_decide_style (GDK_IM_PREEDIT_NONE |  
 *                                  GDK_IM_PREEDIT_NOTHING |
 *                                  GDK_IM_PREEDIT_POSITION |
 *                                  GDK_IM_STATUS_NONE |
 *                                  GDK_IM_STATUS_NOTHING);
 *   ic_attr->client_window = win->window;
 *   ic = gdk_ic_new (ic_attr, GDK_IC_ALL_REQ);
 *   gdk_im_begin (ic, win->window);
 */
    scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
    gtk_container_set_border_width (GTK_CONTAINER (scrolledwindow), -2);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

    gtk_container_add (GTK_CONTAINER (win), scrolledwindow);
    gtk_widget_show (scrolledwindow);

    gtk_widget_push_style (style);
    style = gtk_style_copy (style);
    style->bg[0] = getColorByName (globalColors->rootBg);
    gRootWindow = gtk_fixed_new ();
    gtk_widget_set_style (gRootWindow, style);
    gtk_widget_pop_style ();

    gtk_container_set_border_width (GTK_CONTAINER (gRootWindow), 0);
    gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolledwindow),
	gRootWindow);
    gtk_widget_show (gRootWindow);
    currentParent = gRootWindow;
    pressedKey = 0;
    return 0;
}

int newtFinished(void) {
    free (globalColors);
    return 0;
}

void newtSetColors(struct newtColors colors) {
	globalColors->rootFg = colors.rootBg;
	globalColors->rootBg = colors.rootBg;
	globalColors->borderFg = colors.borderFg;
	globalColors->borderBg = colors.borderBg;
	globalColors->windowFg = colors.windowFg;
	globalColors->windowBg = colors.windowBg;
	globalColors->shadowFg = colors.shadowFg;
	globalColors->shadowBg = colors.shadowBg;
	globalColors->titleFg = colors.titleFg;
	globalColors->titleBg = colors.titleBg;
	globalColors->buttonFg = colors.buttonFg;
	globalColors->buttonBg = colors.buttonBg;
	globalColors->actButtonFg = colors.actButtonFg;
	globalColors->actButtonBg = colors.actButtonBg;
	globalColors->checkboxFg = colors.checkboxFg;
	globalColors->checkboxBg = colors.checkboxBg;
	globalColors->actCheckboxFg = colors.actCheckboxFg;
	globalColors->actCheckboxBg = colors.actCheckboxBg;
	globalColors->entryFg = colors.entryFg;
	globalColors->entryBg = colors.entryBg;
	globalColors->labelFg = colors.labelFg;
	globalColors->labelBg = colors.labelBg;
	globalColors->listboxFg = colors.listboxFg;
	globalColors->listboxBg = colors.listboxBg;
	globalColors->actListboxFg = colors.actListboxFg;
	globalColors->actListboxBg = colors.actListboxBg;
	globalColors->textboxFg = colors.textboxFg;
	globalColors->textboxBg = colors.textboxBg;
	globalColors->actTextboxFg = colors.actTextboxFg;
	globalColors->actTextboxBg = colors.actTextboxBg;
	globalColors->helpLineFg = colors.helpLineFg;
	globalColors->helpLineBg = colors.helpLineBg;
	globalColors->rootTextFg = colors.rootTextFg;
	globalColors->rootTextBg = colors.rootTextBg;
	globalColors->emptyScale = colors.emptyScale;
	globalColors->fullScale = colors.fullScale;
	globalColors->disabledEntryFg = colors.disabledEntryFg;
	globalColors->disabledEntryBg = colors.disabledEntryBg;
	globalColors->compactButtonFg = colors.compactButtonFg;
	globalColors->compactButtonBg = colors.compactButtonBg;
	globalColors->actSelListboxFg = colors.actSelListboxFg;
	globalColors->actSelListboxBg = colors.actSelListboxBg;
	globalColors->selListboxFg = colors.selListboxFg;
	globalColors->selListboxBg = colors.selListboxBg;	
}

int newtGetKey(void) {
    return 0;
}

void newtWaitForKey(void) {
    newtRefresh(); 
    newtClearKeyBuffer();
}

void newtClearKeyBuffer(void) {
}

int newtOpenWindow(int left, int top, int width, int height, 
    const char * title) 
{
    GtkWidget *window, *win;
    GtkStyle *style;

    if (!currentWindow) {
	currentWindow = windowStack;
    } else {
	currentWindow++;
    }

    currentWindow->left = left;
    currentWindow->top = top;
    currentWindow->width = width;
    currentWindow->height = height;
    currentWindow->title = title ? strdup (title) : NULL;
    window = gtk_frame_new (NULL);
    currentWindow->buffer = (short *) NULL;
    currentWindow->widget = window;
    currentParent = window;

    /* creat a background for the frame */
    win = gtk_fixed_new ();
    style = gtk_style_copy (gtk_widget_get_style (win));
    style->bg[0] = getColorByName (globalColors->borderBg);
    gtk_widget_set_style (win, style);
    currentWindow->bg_widget = win;
    gtk_fixed_put (GTK_FIXED (gRootWindow), win,
        FontSizeW * (left-1) - 2, FontSizeH * (top-1));
    gtk_widget_set_usize (win, FontSizeW * (width+1) + 4,
        FontSizeH * (height+1) + 2);
    gtk_widget_show (win);
    
    /* creat the frame */
    style = gtk_style_copy (gtk_widget_get_style (window));
    style->bg[0] = getColorByName (globalColors->windowBg);
    style->fg[0] = getColorByName (globalColors->titleFg);
    gtk_widget_set_style (window, style);
    currentWindow->bg_widget = win;
	
    if (currentWindow->title) {
	gtk_frame_set_label (GTK_FRAME (window), currentWindow->title);
    }
    gtk_frame_set_label_align (GTK_FRAME (window), .5, 0);
    gtk_container_set_border_width (GTK_CONTAINER (window), 0);
    gtk_frame_set_shadow_type (GTK_FRAME (window), GTK_SHADOW_OUT);
    gtk_widget_set_usize (window, FontSizeW * (width+1), 
	FontSizeH * (height+1));
    gtk_fixed_put (GTK_FIXED (win), window, 2, 0); 
    gtk_widget_show (window);

    return 0;
}

int newtCenteredWindow(int width, int height, const char * title) {
    int top, left;

    top = (ScreenRows - height) / 2;

    /* I don't know why, but this seems to look better */ 
    if ((ScreenRows % 2) && (top % 2)) top--;

    left = (ScreenCols - width) / 2;

    newtOpenWindow(left, top, width, height, title);

    return 0;
}

void newtPopWindow(void) {
    int row, col;

    row = col = 0;

    gtk_widget_destroy (currentWindow->widget);
    gtk_widget_destroy (currentWindow->bg_widget);

    free(currentWindow->buffer);
    free(currentWindow->title);

    if (currentWindow == windowStack) { 
	currentWindow = NULL;
        currentParent = gRootWindow;
    } else {
	currentWindow--;
        currentParent = currentWindow->widget;
    }

    newtRefresh(); 
}

void newtGetrc(int * row, int * col) {
   *row = cursorRow;
   *col = cursorCol;
}

void newtGotorc(int newRow, int newCol) {
    if (currentWindow) {
	newRow += currentWindow->top;
	newCol += currentWindow->left;
    }

    cursorRow = newRow;
    cursorCol = newCol;
}

void newtDrawBox(int left, int top, int width, int height, int shadow) {
    if (currentWindow) {
	top += currentWindow->top;
	left += currentWindow->left;
    }

}

void newtClearBox(int left, int top, int width, int height) {
    if (currentWindow) {
	top += currentWindow->top;
	left += currentWindow->left;
    }

}

void newtDelay(int usecs) {
    fd_set set;
    struct timeval tv;

    FD_ZERO(&set);

    tv.tv_sec = usecs / 1000000;
    tv.tv_usec = usecs % 1000000;

    select(0, &set, &set, &set, &tv);
}

struct eventResult newtDefaultEventHandler(newtComponent c,
					   struct event ev) {
    struct eventResult er;

    er.result = ER_IGNORED;
    return er;
}

void newtRedrawHelpLine(void) {
    GtkStyle *style;
    GtkWidget *win;
    if (!currentHelpline) { return; }
    if(!currentHelpWidget) {
        /* creat a background for the help line */
        win = gtk_fixed_new ();
        style = gtk_style_copy (gtk_widget_get_style (win));
        style->bg[0] = getColorByName (globalColors->helpLineBg);
	style->fg[0] = getColorByName (globalColors->helpLineFg);
        gtk_widget_set_style (win, style);
	gtk_widget_set_usize (win, ScreenW  , FontSizeH + 3);
        gtk_fixed_put (GTK_FIXED (gRootWindow), win,
            0, ScreenH - FontSizeH - 3);
        gtk_widget_show (win);

        currentHelpWidget = gtk_statusbar_new ();
	gtk_widget_set_usize (currentHelpWidget, ScreenW , FontSizeH + 3);
	gtk_fixed_put (GTK_FIXED (win), currentHelpWidget, 1, 0);
	gtk_widget_set_style (
		(GTK_STATUSBAR(currentHelpWidget)->label), style);
	gtk_widget_show (currentHelpWidget);
    }
    gtk_statusbar_push (GTK_STATUSBAR (currentHelpWidget), 1, *currentHelpline);
}

void newtPushHelpLine(const char * text) {
    if (!text)
	text = defaultHelpLine;
    
    if (currentHelpline)
	(*(++currentHelpline)) = strdup(text);
    else {
	currentHelpline = helplineStack;
	*currentHelpline = strdup(text);
    }

    newtRedrawHelpLine();
}

void newtPopHelpLine(void) {
    if (!currentHelpline) return;

    free(*currentHelpline);
    if (currentHelpline == helplineStack)
	currentHelpline = NULL;
    else
	currentHelpline--;

    newtRedrawHelpLine();

}

void newtDrawRootText(int col, int row, const char * text) {
    GtkWidget *w;
    GtkStyle *style;

    if (col < 0) {
        col = ScreenCols + col;
    }

    if (row < 0) {
        row = ScreenRows + row;
    }

    w = gtk_label_new ((char *)text);
    style = gtk_style_copy (gtk_widget_get_style (w));
    style->fg[0] = getColorByName (globalColors->rootTextFg);
    gtk_widget_set_style (w, style);
    gtk_widget_show (w);
    gtk_fixed_put (GTK_FIXED (gRootWindow), w, 
	FontSizeW * col , 
	FontSizeH * row );
}

int newtSetFlags(int oldFlags, int newFlags, enum newtFlagsSense sense) {
    switch (sense) {
      case NEWT_FLAGS_SET:
	return oldFlags | newFlags;

      case NEWT_FLAGS_RESET:
	return oldFlags & (~newFlags); 

      case NEWT_FLAGS_TOGGLE:
	return oldFlags ^ newFlags;

      default:
	return oldFlags;
    }
}

void newtBell(void)
{
}

void newtGetScreenSize(int * cols, int * rows) {
    if (rows) *rows = ScreenRows;
    if (cols) *cols = ScreenCols;
}

void newtDefaultPlaceHandler(newtComponent c, int newLeft, int newTop) {
    c->left = newLeft;
    c->top = newTop;
}

void newtDefaultMappedHandler(newtComponent c, int isMapped) {
    c->isMapped = isMapped;
}
